import numpy as np
from astropy.modeling import models
import matplotlib.pyplot as plt
import gsaoi_static_distortion_info as gsdi

def default_static_model(extver, date=None, linear=False):
    """
    Calculate the appropriate static distortion model for this array
    using the coefficients in the lookup table
    Input: extver = array ID (1-4)
           date: date of observations
           linear: return linear (as well as non-linear) component?
    Output: list of models to accomplish the mapping
    """
    
    # This is here in case the correction changes with time
    date_dict = gsdi.STATIC_CORRECTIONS_DATE_MAPPING
    for (start_date, end_date) in date_dict.keys():
        return_key = (start_date, end_date)
        if end_date is None:
            if start_date is None or date is None:
                break
            elif date >= start_date:
                break
        elif date >= start_date and date < end_date:
            break
    else:
        raise IOError("Cannot determine which static correction to use")
    
    # Retrieve the static_correction_dictionaries
    top_level_dict = date_dict[return_key]
    full_forward_dict = getattr(gsdi, top_level_dict[gsdi.MODEL])
    full_reverse_dict = getattr(gsdi, top_level_dict[gsdi.INVERSE])
    cor_offset = getattr(gsdi, top_level_dict[gsdi.CENTRE_OF_ROTATION_OFFSET])

    # Correction for trimming (4 pixels) plus 1-indexing. Lazily hardcoded.
    indexing_transform = (models.Shift(cor_offset['x']+5) &
                models.Shift(cor_offset['y']+5)).rename('IndexingTransform')

    # Pix2TNX projection transform
    forward_model = []
    use_corrections = full_forward_dict[gsdi.CORRECTION_ORDER] if linear else ["non-linear"]
    for model in use_corrections:
        model_dict = full_forward_dict[model][gsdi.ARRAY_NAMES["A"+str(extver)]]
        model_dict.update({'model': full_forward_dict[model][gsdi.MODEL]})
        forward_model.append(model_dict)
    reverse_model = []
    for model in full_reverse_dict[gsdi.CORRECTION_ORDER]:
        model_dict = full_reverse_dict[model][gsdi.ARRAY_NAMES["A"+str(extver)]]
        model_dict.update({'model': full_reverse_dict[model][gsdi.MODEL]})
        reverse_model.append(model_dict)
    pix2sky_tnx_model = make_tnx_model(forward_model, reverse_model)
    
    return [indexing_transform, pix2sky_tnx_model]

def make_tnx_model(forward_model, reverse_model):
    """
    Return an instance of a Pix2Sky_TNX transform and its inverse
    Inputs: forward_model = list of dicts describing forward transformation
            reverse_model = list of dicts describing reverse transformation
            These dicts need to have the following keys:
              model = name of model (only Polynomial2D, Legendre2D,
                                          Chebyshev2D are currently supported)
              parameters = dict with 'x' and 'y' keys listing the parameters
                           for the model for each ordinate
    """
    forward_transform = None
    for direction in [forward_model, reverse_model]:
        for counter, model in enumerate(direction):
            model_type = model['model']
            transform = {}
            for ordinate in ['x','y']:
                # TODO Act reasonably if unacceptable model is passed

                # Get the degree of the model (needed to instantiate)
                # Parameter (coefficient) names are of form cm_n, where m
                # is the degree in x and n is the degree in y
                max_x_power = np.max([int(key[1])
                                      for key in model['parameters'][ordinate]])
                max_y_power = np.max([int(key[3])
                                      for key in model['parameters'][ordinate]])
                if model_type == 'Polynomial2D':
                    t = models.Polynomial2D(degree=max(max_x_power,max_y_power))
                elif model_type in ['Chebyshev2D','Legendre2D']:
                    t = getattr(models, model_type)(x_degree=max_x_power,
                                                    y_degree=max_y_power)
                else:
                    pass
                # Set model parameters
                parameters = model['parameters'][ordinate]
                for param_name, param_value in parameters.items():
                    setattr(t, param_name, param_value)
                transform.update({ordinate: t})
    
            # Construct a model that takes (x,y) as inputs and outputs (x',y')
            # Since both x' and y' need (x,y) as inputs, we must duplicate inputs
            m = models.Mapping((0,1,0,1)) | (transform['x'] & transform['y'])

            if counter == 0:
                direction_model = m
            else:
                direction_model = direction_model + m

        # Create the transform / inverse
        if forward_transform:
            forward_transform.inverse = direction_model
        else:
            forward_transform = direction_model

    return forward_transform.rename('Pix2Pix_TNX')

# Create grid of fiducial points
xraw, yraw = np.mgrid[11:2040:32,11:2040:32]

for extver in range(1,5):
    indexing_transform, tnx_model = default_static_model(extver, linear=False)
    # Determine centre of array in transformed coordinates
    x, y = indexing_transform(1020, 1020)
    xcent, ycent = tnx_model(x, y)
    
    # Determine transformed coordinates for grid
    xin, yin = indexing_transform(xraw, yraw)
    xout, yout = tnx_model(xin, yin)
    
    # Calculate offsets in pixel space (assume 0.0195"/pixel)
    dx = (xout-xcent)/0.0195+1020 - xraw
    dy = (yout-ycent)/0.0195+1020 - yraw

    # Plot something    
    plt.quiver(xraw,yraw, dx,dy)
    plt.xlim(0,2040)
    plt.ylim(0,2040)
    plt.show()